<?php

class WPshiftBackup
{
    private $backup_dir;
    public $zip_file;
    private $db_file;
    private $zip;
    private $progress_file;

    public function __construct() {
        $upload_dir = wp_upload_dir();
        $this->backup_dir = $upload_dir['basedir'] . '/wpshift-backups';
        $this->zip_file = $this->backup_dir . '/site-backup.zip';
        $this->db_file = $this->backup_dir . '/db.sql';
        $this->progress_file = $this->backup_dir . '/progress.json';

        if (!file_exists($this->backup_dir)) {
            wp_mkdir_p($this->backup_dir);
        }

        $this->zip = new ZipArchive();

        if (file_exists($this->progress_file)) {
            $this->progress = json_decode(file_get_contents($this->progress_file), true);
        } else {
            $this->progress = ['files' => [], 'db' => []];
        }
    }

    public function is_backup_complete() {
        return file_exists($this->zip_file) && isset($this->progress['files']['done']) && $this->progress['files']['done'] && isset($this->progress['db']['done']) && $this->progress['db']['done'];
    }

    public function create_backup() {
        $this->backup_files();
        $this->backup_database();
    }

    public function backup_files() {
        $file_batch_size = $this->calculate_batch_size();

        if (!isset($this->progress['files']['done']) || !$this->progress['files']['done']) {
            if ($this->zip->open($this->zip_file, ZipArchive::CREATE) !== true) {
                return new WP_Error('zip_creation_failed', 'Could not create zip file.');
            }

            $files = $this->get_all_files(ABSPATH);
            $start = isset($this->progress['files']['start']) ? $this->progress['files']['start'] : 0;

            // Normalize ABSPATH
            $abs_path = realpath(ABSPATH);
            if (substr($abs_path, -1) !== DIRECTORY_SEPARATOR) {
                $abs_path .= DIRECTORY_SEPARATOR;
            }

            for ($i = $start; $i < $start + $file_batch_size && $i < count($files); $i++) {
                $file = $files[$i];
                if (is_file($file)) {
                    // Normalize file path
                    $file_path = realpath($file);

                    // Create relative path
                    $relative_path = str_replace($abs_path, '', $file_path);

                    // Remove leading slashes from relative path
                    $relative_path = ltrim($relative_path, '/\\');

                    // Add file to zip archive
                    $this->zip->addFile($file_path, $relative_path);
                }
            }

            $this->progress['files']['start'] = $i;
            if ($i >= count($files)) {
                $this->progress['files']['done'] = true;
                // Delay closing the zip until we add the database backup
            }

            file_put_contents($this->progress_file, json_encode($this->progress));

            if ($i < count($files)) {
                return ['status' => 'partial', 'message' => 'File backup in progress...', 'progress' => $i / count($files) * 100];
            }
        }

        return ['status' => 'done', 'message' => 'Files backup completed.'];
    }

    public function backup_database() {
        global $wpdb;

        $db_batch_size = $this->calculate_batch_size('database');


        if (!isset($this->progress['db']['done']) || !$this->progress['db']['done']) {
            $tables = $wpdb->get_results('SHOW TABLES', ARRAY_N);
            $start = isset($this->progress['db']['start']) ? $this->progress['db']['start'] : 0;

            for ($i = $start; $i < $start + $db_batch_size && $i < count($tables); $i++) {
                $table_name = $tables[$i][0];
                $create_table_stmt = $wpdb->get_row("SHOW CREATE TABLE $table_name", ARRAY_N);
                file_put_contents($this->db_file, $create_table_stmt[1] . ";\n\n", FILE_APPEND);

                $rows = $wpdb->get_results("SELECT * FROM $table_name", ARRAY_A);
                foreach ($rows as $row) {
                    $values = array_map([$wpdb, 'escape'], array_values($row));
                    $values = "'" . implode("','", $values) . "'";
                    $sql = "INSERT INTO $table_name VALUES ($values);\n";
                    file_put_contents($this->db_file, $sql, FILE_APPEND);
                }
                file_put_contents($this->db_file, "\n", FILE_APPEND);
            }

            $this->progress['db']['start'] = $i;
            if ($i >= count($tables)) {
                $this->progress['db']['done'] = true;

                // Add the database backup file to the ZIP
                if ($this->zip->open($this->zip_file) === true) {
                    $this->zip->addFile($this->db_file, 'db.sql');
                    $this->zip->close();
                }
            }

            file_put_contents($this->progress_file, json_encode($this->progress));

            if ($i < count($tables)) {
                return ['status' => 'partial', 'message' => 'Database backup in progress...', 'progress' => $i / count($tables) * 100];
            }
        }

        return ['status' => 'done', 'message' => 'Database backup completed and included in ZIP.'];
    }

    public function regenerate_backup() {
        // Delete old backups
        $this->cleanup();

        // Start a new backup
        $this->progress = ['files' => [], 'db' => []];
        $this->create_backup();
    }

    public function cleanup() {
        // Delete the ZIP, SQL, and progress files if they exist
        if (file_exists($this->zip_file)) {
            unlink($this->zip_file);
        }

        if (file_exists($this->db_file)) {
            unlink($this->db_file);
        }

        if (file_exists($this->progress_file)) {
            unlink($this->progress_file);
        }
    }

    private function get_all_files($dir) {
        $files = [];
        $items = scandir($dir);

        foreach ($items as $item) {
            if ($item == '.' || $item == '..') {
                continue;
            }

            $path = $dir . '/' . $item;
            if (is_dir($path)) {
                $files = array_merge($files, $this->get_all_files($path));
            } else {
                $files[] = $path;
            }
        }

        return $files;
    }

    private function calculate_batch_size($context = 'files', $target_time = 1.0) {
        // $context can be 'files' or 'database'
        // $target_time is the maximum time (in seconds) a batch should take

        $start_time = microtime(true);
        $batch_size = 10; // Start with a small batch size for testing

        if ($context === 'files') {
            // Test with a small number of files
            $files = $this->get_all_files(ABSPATH);
            $test_files = array_slice($files, 0, $batch_size);
            foreach ($test_files as $file) {
                if (is_file($file)) {
                    $file_path = realpath($file);
                    $this->zip->addFile($file_path, basename($file_path)); // Mock file addition
                }
            }
        } elseif ($context === 'database') {
            global $wpdb;
            // Test with a small number of rows
            $tables = $wpdb->get_results('SHOW TABLES', ARRAY_N);
            $table_name = $tables[0][0]; // Test with the first table
            $rows = $wpdb->get_results("SELECT * FROM $table_name LIMIT $batch_size", ARRAY_A);
            foreach ($rows as $row) {
                // Mock processing rows
                json_encode($row); // Simulate some processing
            }
        }

        $elapsed_time = microtime(true) - $start_time;

        // Adjust batch size based on elapsed time
        if ($elapsed_time > 0) {
            $optimal_batch_size = (int)($batch_size / $elapsed_time * $target_time);
        } else {
            $optimal_batch_size = $batch_size * 2; // Arbitrary doubling if time is too fast to measure
        }

        // Cap the batch size to a reasonable maximum to avoid resource exhaustion
        return min($optimal_batch_size, 1000);
    }
}
